@@ -1,35 +1,71 @@
+use v5.16; # from dependencies, so go for it.
 package CPAN::Mini::Inject;
 
 use strict;
 use warnings;
 
-use CPAN::Checksums qw( updatedir );
+use CPAN::Checksums 2.13;
 use CPAN::Mini;
 use CPAN::Mini::Inject::Config;
 use Carp;
 use Compress::Zlib;
-use Env;
 use File::Basename;
 use File::Copy;
 use File::Path qw( make_path );
 use File::Spec;
+use File::Spec::Functions;
 use LWP::Simple;
 use Dist::Metadata ();
 
+BEGIN {
+  use version 0.9915;
+  use CPAN::Meta::Converter;
+
+  # This is here because the CPAN::Meta package has not been updated
+  # since 2016 and it's unlikely that they'd accept a patch for this.
+  # see https://github.com/briandfoy/cpan-mini-inject/issues/11
+  # and https://github.com/Perl-Toolchain-Gang/CPAN-Meta#138
+  package # hide from PAUSE
+    CPAN::Meta::Converter;
+
+  no warnings qw(redefine);
+
+  # lifted from CPAN::Meta::Converter
+  # https://fastapi.metacpan.org/source/DAGOLDEN/CPAN-Meta-2.150010/lib/CPAN/Meta/Converter.pm
+  sub _clean_version {
+    my ($element) = @_;
+    return 0 if ! defined $element;
+
+    $element =~ s{^\s*}{};
+    $element =~ s{\s*$}{};
+    $element =~ s{^\.}{0.};
+
+    return 0 if ! length $element;
+    return 0 if ( $element eq 'undef' || $element eq '<undef>' );
+    my $v = eval { version->parse($element) };
+    # XXX check defined $v and not just $v because version objects leak memory
+    # in boolean context -- dagolden, 2012-02-03
+    if ( defined $v ) {
+    return _is_qv($v) ? $v->stringify : $element;
+    }
+    else {
+    return 0;
+    }
+  }
+}
+
+=encoding utf8
+
 =head1 NAME
 
 CPAN::Mini::Inject - Inject modules into a CPAN::Mini mirror.
 
-=head1 VERSION
-
-Version 0.33
-
 =cut
 
-our $VERSION = '0.33';
+our $VERSION = '1.008';
 our @ISA     = qw( CPAN::Mini );
 
-=head1 Synopsis
+=head1 SYNOPSIS
 
 If you're not going to customize the way CPAN::Mini::Inject works you
 probably want to look at the L<mcpani> command instead.
@@ -39,10 +75,12 @@ probably want to look at the L<mcpani> command instead.
     $mcpi=CPAN::Mini::Inject->new;
     $mcpi->parsecfg('t/.mcpani/config');
 
-    $mcpi->add( module   => 'CPAN::Mini::Inject',
-                authorid => 'SSORICHE',
-                version  => ' 0.01',
-                file     => 'mymodules/CPAN-Mini-Inject-0.01.tar.gz' )
+    $mcpi->add(
+      module   => 'CPAN::Mini::Inject',
+    authorid => 'SSORICHE',
+    version  => ' 0.01',
+    file     => 'mymodules/CPAN-Mini-Inject-0.01.tar.gz'
+  );
 
     $mcpi->writelist;
     $mcpi->update_mirror;
@@ -50,9 +88,15 @@ probably want to look at the L<mcpani> command instead.
 
 =head1 DESCRIPTION
 
-CPAN::Mini::Inject uses CPAN::Mini to build or update a local CPAN mirror
-then adds modules from your repository to it, allowing the inclusion
-of private modules in a minimal CPAN mirror.
+CPAN::Mini::Inject uses CPAN::Mini to build or update a I<local> CPAN mirror
+from a I<remote> one.  It adds two extra features:
+
+1. an additional I<repository> of distribution files and related information
+(author and module versions), separate from the local and remote mirrors, to
+which you can add your own distribution files.
+
+2. the ability to I<inject> the distribution files from your I<repository>
+into a I<local> CPAN mirror.
 
 =head1 METHODS
 
@@ -68,53 +112,69 @@ A C<CPAN::Mini::Inject> ISA L<CPAN::Mini>. Refer to the
 L<documentation|CPAN::Mini> for that module for details of the interface
 C<CPAN::Mini::Inject> inherits from it.
 
-=head2 C<new>
+=over 4
+
+=item C<new>
 
 Create a new CPAN::Mini::Inject object.
 
 =cut
 
 sub new {
-  return bless
-   { config_class => 'CPAN::Mini::Inject::Config' },
-   $_[0];
+  my( $class, %args ) = @_;
+  my %defaults = (
+    config_class => $class->default_config_class,
+  );
+  my %allowed = map {
+    $_, 1
+  } qw(config_class);
+
+  my %filtered =
+    map { ($_, $args{$_}) }
+    grep { exists $allowed{$_} }
+    keys %args;
+
+
+  my %obj = ( %defaults, %filtered );
+  return bless \%obj, $class;
 }
 
-=head2 C<< config_class( [CLASS] ) >>
+=item C<< config_class( [CLASS] ) >>
 
-Returns the name of the class handling the configuration.
-
-With an argument, it sets the name of the class to handle
-the config. To use that, you'll have to call it before you
-load the configuration.
+Returns the name of the class used to handle the configuration. Also
+see C<default_config_class>.
 
 =cut
 
 sub config_class {
   my $self = shift;
-
   if ( @_ ) { $self->{config_class} = shift }
-
   $self->{config_class};
 }
 
-=head2 C<< config >>
+=item C<< config( [HASHREF] ) >>
 
-Returns the configuration object. This object should be from
-the class returned by C<config_class> unless you've done something
-weird.
+With a hashref argument, sets the config data.
+
+Returns the current configuration hash.
 
 =cut
 
 sub config {
   my $self = shift;
-
   if ( @_ ) { $self->{config} = shift }
-
   $self->{config};
 }
 
-=head2 C<< loadcfg( [FILENAME] ) >>
+=item C<< default_config_class >>
+
+=cut
+
+sub default_config_class {
+  'CPAN::Mini::Inject::Config'
+}
+
+=item C<< loadcfg( [FILENAME] ) >>
 
 
 This is a bridge to CPAN::Mini::Inject::Config's loadconfig. It sets the
@@ -134,7 +194,7 @@ sub loadcfg {
   return $self;
 }
 
-=head2 C<< parsecfg() >>
+=item C<< parsecfg() >>
 
 This is a bridge to CPAN::Mini::Inject::Config's parseconfig.
 
@@ -152,10 +212,11 @@ sub parsecfg {
   return $self;
 }
 
-=head2 C<< site( [SITE] ) >>
+=item C<< site( [SITE] ) >>
 
-Returns the CPAN site that CPAN::Mini::Inject chose from the
-list specified in the C<remote> directive.
+With an argument, set the site to use to contact CPAN. Returns the
+site setting, or, if the site has not be set (or was set to undef),
+returns the empty string.
 
 =cut
 
@@ -165,14 +226,15 @@ sub site {
 
   if ( @_ ) { $self->{site} = shift }
 
-  $self->{site} || '';
+  $self->{site} // '';
 }
 
-=head2 C<testremote>
+=item C<testremote>
 
-Test each site listed in the remote parameter of the config file by performing
-a get on each site in order for authors/01mailrc.txt.gz. The first site to
-respond successfully is set as the instance variable site.
+Test each site listed in the remote parameter of the config file by
+performing a get on each site in order for authors/01mailrc.txt.gz.
+The first site to respond successfully is set as the instance variable
+site.
 
  print "$mcpi->{site}\n"; # ftp://ftp.cpan.org/pub/CPAN
 
@@ -207,7 +269,7 @@ sub testremote {
   return $self;
 }
 
-=head2 C<update_mirror>
+=item C<update_mirror>
 
 This is a subclass of CPAN::Mini.
 
@@ -222,9 +284,15 @@ sub update_mirror {
 
   $ENV{FTP_PASSIVE} = 1 if $self->config->get( 'passive' );
 
-  $options{local}     ||= $self->config->get( 'local' );
-  $options{trace}     ||= 0;
-  $options{skip_perl} ||= $self->config->get( 'perl' ) || 1;
+  $options{local}        ||= $self->config->get( 'local' );
+  $options{trace}        ||= 0;
+  $options{skip_perl}    ||= $self->config->get( 'perl' ) || 1;
+  $options{skip_cleanup} ||= $self->config->get( 'skip_cleanup' ) || 0;
+
+  # module_filters, log_level, and force
+  my @extra = grep { defined $self->config->get($_) } qw(module_filters log_level force);
+
+  $options{$_} = $self->config->get($_) for @extra;
 
   $self->testremote( $options{trace} )
    unless ( $self->site || $options{remote} );
@@ -236,52 +304,55 @@ sub update_mirror {
   CPAN::Mini->update_mirror( %options );
 }
 
-=head2 C<add>
+=item C<add>
 
-Add a new module to the repository. The add method copies the module
-file into the repository with the same structure as a CPAN site. For
-example CPAN-Mini-Inject-0.01.tar.gz is copied to
-MYCPAN/authors/id/S/SS/SSORICHE. add creates the required directory
-structure below the repository.
+Add a new distribution to the repository. The C<add> method copies the
+distribution file into the repository with the same structure as a
+CPAN site. For example, F<CPAN-Mini-Inject-0.01.tar.gz> with author
+C<SSORICHE> is copied to F<MYCPAN/authors/id/S/SS/SSORICHE>. add
+creates the required directory structure below the repository.
 
 Packages found in the distribution will be added to the module list
-(for example both C<CPAN::Mini::Inject> and C<CPAN::Mini::Inject::Config>
-will be added to the F<modules/02packages.details.txt.gz> file).
+(for example both C<CPAN::Mini::Inject> and
+C<CPAN::Mini::Inject::Config> will be added to the
+F<modules/02packages.details.txt.gz> file).
 
-Packages will be looked for in the C<provides> key of the META file if present,
-otherwise the files in the dist will be searched.
-See L<Dist::Metadata> for more information.
+Packages will be looked for in the C<provides> key of the META file if
+present, otherwise the files in the dist will be searched. See
+L<Dist::Metadata> for more information.
 
 =over 4
 
 =item * module
 
-The name of the module to add.
-The distribution file will be searched for modules
-but you can specify the main one explicitly.
+(optional) The package name of the module to add. The distribution
+file will be searched for modules but you can specify the main one
+explicitly.
 
 =item * authorid
 
-CPAN author id. This does not have to be a real author id.
+(required) The CPAN ID of the module's author. Since this isn't
+actually CPAN, the ID does not need to exist on CPAN. Typically, this
+ID uses C<[A-Z]> and is three to ten letters. This is not enforced,
+but other CPAN tools may not like other sorts of names.
 
 =item * version
 
-The modules version number.
-Module names and versions will be determined,
-but you can specify one explicitly.
+(optional) The module's version number. If you don't specify this.
+C<add> will try to extract it from the distribution.
 
 =item * file
 
-The tar.gz of the module.
+(required) The path to the distribution file.
 
 =back
 
-=head3 Example
-
-  add( module => 'Module::Name',
-       authorid => 'AUTHOR',
-       version => 0.01,
-       file => './Module-Name-0.01.tar.gz' );
+  $mcpani->add(
+    module   => 'Module::Name',
+  authorid => 'SOMEAUTHOR',
+  version  => 0.01,
+  file     => './Module-Name-0.01.tar.gz'
+  );
 
 =cut
 
@@ -304,10 +375,10 @@ sub add {
 
   # attempt to guess module and version
   my $distmeta = Dist::Metadata->new( file => $options{file} );
-  my $packages = $distmeta->package_versions;
 
-  # include passed in module and version (prefer discovered version)
-  if ( $options{module} ) {
+  my $packages = $distmeta->package_versions;
+  # include passed in module and version (prefer the declared version)
+  if ( $options{module} and $options{version} ) {
     $packages->{ $options{module} } ||= $options{version};
   }
 
@@ -362,7 +433,7 @@ sub add {
   return $self;
 }
 
-=head2 C<added_modules>
+=item C<added_modules>
 
 Returns a list of hash references describing the modules added by this instance.
 Each hashref will contain C<file>, C<authorid>, and C<modules>.
@@ -381,7 +452,7 @@ sub added_modules {
   return @{ $self->{added_modules} ||= [] };
 }
 
-=head2 C<inject>
+=item C<inject>
 
 Insert modules from the repository into the local CPAN::Mini mirror. inject
 copies each module into the appropriate directory in the CPAN::Mini mirror
@@ -403,12 +474,17 @@ sub inject {
 
   my %updatedir;
   my %already_injected;
+  my %report;
   for my $modline ( @{ $self->{modulelist} } ) {
     my ( $module, $version, $file ) = split( /\s+/, $modline );
 
-    next if $already_injected{$file}++;
-
     my $target = $self->config->get( 'local' ) . '/authors/id/' . $file;
+
+    # collect all modules of a target/file
+    # needed for report
+    push @{ $report{$target} }, $module;
+    next if $already_injected{$module}++;
+
     my $source
      = $self->config->get( 'repository' ) . '/authors/id/' . $file;
 
@@ -420,14 +496,27 @@ sub inject {
      or croak "Copy $source to $tdir failed: $!";
 
     $self->_updperms( $target );
-    print "$target ... injected $module\n" if $verbose;
+  }
+
+  # if verbose report target file and the modules it contains
+  if ( $verbose ) {
+    for my $target (keys %report) {
+        my $target_str = "$target ... injected modules : ";
+        my $fmt = '%' . length($target_str) . "s%s\n";
+        my @modules = @{ $report{$target} };
+        printf $fmt, $target_str, shift @modules;    # first line with target
+        for my $module ( @modules ) {                # rest only the module
+            printf $fmt, '', $module;
+        }
+    }
   }
 
   for my $dir ( keys( %updatedir ) ) {
-    my $authdir = $self->config->get( 'local' ) . "/authors/id/$dir";
+    my $root    = catfile( $self->config->get( 'local' ), qw(authors id) );
+    my $authdir = catfile( $root, $dir );
 
-    updatedir( $authdir );
-    $self->_updperms( "$authdir/CHECKSUMS" );
+    CPAN::Checksums::updatedir( $authdir, $root );
+    $self->_updperms( catfile($authdir, 'CHECKSUMS') );
   }
 
   $self->updpackages;
@@ -436,7 +525,7 @@ sub inject {
   return $self;
 }
 
-=head2 C<updpackages>
+=item C<updpackages>
 
 Update the CPAN::Mini mirror's modules/02packages.details.txt.gz with the
 injected module information.
@@ -457,10 +546,10 @@ sub updpackages {
     $packages{$pkg} = $line;
   };
 
-  $self->_writepkgs( [ sort values %packages ] );
+  $self->_writepkgs( [ sort { lc $a cmp lc $b } values %packages ] );
 }
 
-=head2 C<updauthors>
+=item C<updauthors>
 
 Update the CPAN::Mini mirror's authors/01mailrc.txt.gz with
 stub information should the author not actually exist on CPAN
@@ -481,10 +570,7 @@ sub updauthors {
   AUTHOR:
   for my $modline ( @{ $self->{modulelist} } ) {
     my ( $module, $version, $file ) = split( /\s+/, $modline );
-
-    # extract the author from the path
-    my @dirs = File::Spec->splitdir( $file );
-    my $author = $dirs[2];
+    my $author = (File::Spec->splitdir( $file ))[2];
 
     next AUTHOR if defined $author_ids_in_repo{$author};
     next AUTHOR if defined $authors_added{$author};
@@ -499,7 +585,7 @@ sub updauthors {
 
 }
 
-=head2 C<readlist>
+=item C<readlist>
 
 Load the repository's modulelist.
 
@@ -529,7 +615,7 @@ sub readlist {
   return $self;
 }
 
-=head2 C<writelist>
+=item C<writelist>
 
 Write to the repository modulelist.
 
@@ -709,33 +795,30 @@ sub _fmtdate {
   return "$date[0], $date[2] $date[1] $date[4] $date[3] GMT";
 }
 
-=head1 See Also
+=back
+
+=head1 SEE ALSO
 
 L<CPAN::Mini>
 
-=head1 Current Maintainer
-
-Christian Walde C<< <walde.christian@googlemail.com> >>
-
 =head1 Original Author
 
 Shawn Sorichetti, C<< <ssoriche@cpan.org> >>
 
-=head1 Acknowledgements
+=head1 ACKNOWLEDGEMENTS
 
 Special thanks to David Bartle, for bringing this module up
 to date, and resolving the reported bugs.
 
 Thanks to Jozef Kutej <jozef@kutej.net> for numerous patches.
 
-=head1 Bugs
+=head1 BUGS
 
-Please report any bugs or feature requests to
-C<bug-cpan-mini-inject@rt.cpan.org>, or through the web interface at
-L<http://rt.cpan.org>.  I will be notified, and then you'll automatically
-be notified of progress on your bug as I make changes.
+Report issues to the GitHub queue at
 
-=head1 Copyright & License
+  https://github.com/briandfoy/cpan-mini-inject/issues
+
+=head1 COPYRIGHT AND LICENSE
 
 Copyright 2008-2009 Shawn Sorichetti, Andy Armstrong, All Rights Reserved.
 
